/* FUSH - $VERSION$
This file is part of FUSH (Free Unix SHell, Fine User SHell, or FU shell)

   FUSH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   FUSH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Foobar; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

* Author:  Cory Marsh
* Contact: cmarsh@idahopower.com
*/
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif
#include "fush.h"
#include <signal.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/resource.h>

#ifdef HAVE_LIBREADLINE
#include <readline/history.h>
#include <readline/readline.h>
#endif


/* GLOBAL LISTS */
List g_procs;
List g_exec;
List g_db;
List g_alias;
List g_conf;
List g_seconf;
List g_replace;
List g_setenv;
List g_owner;
List g_group;

void catch_io( int signo ) {
  printf( "there is data to read\n" );
  die( "signal" );
}

/* Wed Nov 12 21:39:07 PST 2003
 * wait on finished child processes */
void catch_chld( int signo ) {
  int status;
  char buffer[ BUFFLARG ];
  char *value;
  pid_t pid;
  back_cmd *cmd = NULL;

  if ( ( pid = waitpid( 0, &status, WNOHANG ) ) < 0 && errno != ECHILD )
    perror( "waitpid(3)" );

  /* if this finished process was a background task, do some accounting */
  cmd = find_proc( &g_procs, pid );

  /* print the finished process to the info buffer, NULL ptr accounting */
  if ( cmd != NULL ) {
    value = fu_get_conf( &g_conf, "notify", 0 );
    if( value )
      snprintf( buffer, BUFFLARG, "%s[%d] - Done PID: [%d] Command : [%s]\n",
                value, cmd->c_num, (int)cmd->pid, cmd->command );
    else
      snprintf( buffer, BUFFLARG, "[%d] - Done PID: [%d] Command : [%s]\n",
                cmd->c_num, (int)cmd->pid, cmd->command );
    fu_add_value( &g_conf, "shell_notify", buffer );
    free( cmd );
  }

  if( status != 0 ) {
    snprintf( buffer, BUFFLARG, "%d", status ); 
    fu_add_value( &g_conf, "status", buffer );
  }
  else
    fu_get_conf( &g_conf, "status", 1 );

  return ;
}

void catch_int( int signo ) {
  printf("cought int\n");
  // signal( SIGINT,  &catch_int );
  //process_cmds();
}

/* Sat Nov  8 23:03:36 PST 2003
 * fork and execute a command searching path. support for background procs */
int fu_exec( char* command, char **exec_args, int outfd, int infd,
             int flag ) {
  pid_t pid;
  int status;
  char buffer[ BUFFMEDM ];
  char realpath[ MAXPATHLEN ];
  char *cmdname;
  int  iconf=0; 
  back_cmd *cmd;
  back_cmd *rtr;


  /* avoid obvious errors, and don't execute if this flag is set */
  if ( command == NULL || ( flag & F_NOEXEC ) )
    return ( -1 );

  /* bail out early on error conditions ( > max background process ) */
  iconf = atoi( fu_get_conf( &g_seconf, "fu_max_background", 0 ) );
  if ( ( flag & F_BACK ) && list_size( &g_procs ) >= iconf ) {
    syndenied( "> max background processes" );
    return ( -1 );
  }

  /* substitute commands, and get real paths */
  if ( ( cmdname = resolve_path( command, realpath, MAXPATHLEN ) ) == NULL ) {
    fprintf( stderr, "%s: Command not found\n", command );
    return ( -1 );
  }

  /* cmdname no longer needs free, will be freed by freeglob() */
  /* MEMORY LEAK CHECK */
  //free( exec_args[0] );
  exec_args[0] = cmdname;

#ifdef DEBUG
  printf( "command [%s] input [%d]\n", command, infd );
  printf( "path: [%s] name: [%s]\n", realpath, cmdname );
#endif

  /* clean the environment! */
  fu_clean_env( &g_setenv );

  pid = fork();
  /* child process */
  if ( pid == 0 ) {

    /* dup stdout to our pipe if it connects to not stdout */
    if ( outfd != 1 ) {
      if ( dup2( outfd, 1 ) == -1 )
        die( "can't redirect stdout" );
    }

    /* dup stdin to our pipe if it connects to not stdin */
    if ( infd != 0 ) {
      if ( dup2( infd, 0 ) == -1 )
        die( "can't redirect stdin" );
    }

    /* signal handling */
    signal( SIGINT, SIG_DFL );
    if ( flag & F_NOHUP )
      signal( SIGHUP, SIG_IGN );
    if ( flag & F_NOKILL )
      signal( SIGKILL, SIG_IGN );
  
    if ( flag & F_BACK ) {
      iconf = atoi( fu_get_conf( &g_seconf, "fu_background_nice", 0 ) );
      setpriority( PRIO_PROCESS, 0, iconf );
    }

    /* don't exec stuff that doesnt conform to our policy */
    status = can_exec( &g_exec, realpath );

    if ( status == 0 ) {
      snprintf( buffer, BUFFMEDM, "exec %s", command );
      syndenied( buffer );
      return ( -1 );
    }
    else if ( status == 2 ) {
      fprintf( stderr, "%s: Command not found\n", command );
      return ( -1 );
    }

    /* exec! */
    if ( execv( realpath, exec_args ) == -1 )
      perror( cmdname );

    /* control not reached */
    fu_exit( -1 );
  }

  /* parent process */
  else if ( pid > 0 ) {

    /* wait for the child to finish before restoring control
    * this always returns -1 becuase the child is caught by signal */
    if ( ! ( flag & F_BACK ) ) {
      if( ( pid = waitpid( pid, &status, 0 ) < 0 ) && errno != ECHILD )
        perror( "waitpid(3)" );
    }

    /* add child to list of background processes and restore control */
    else if ( ( flag & F_BACK ) ) {

      /* runing processes accounting */
      if ( ( cmd = ( back_cmd * ) malloc( sizeof( back_cmd ) ) ) == NULL )
        die( "malloc()" );
      bzero( cmd, sizeof( back_cmd ) );

      snprintf( cmd->command, MAXPATHLEN, "%s", cmdname );

      if ( ( rtr = find_proc( &g_procs, 0 ) ) == NULL )
        cmd->c_num = 1;
      else
        cmd->c_num = rtr->c_num + 1;

      cmd->pid = pid;

      /* add this process to our list of background procs */
      list_ins_next( &g_procs, list_tail( &g_procs ), cmd );
    }

    /* close output pipe after command has run */
    if ( outfd > 2 && close( outfd ) == -1 )
      perror( "can't close output pipe!" );

    /* close input pipe after command has run -IF- input is not STDIN */
    if ( infd > 2 && close( infd ) == -1 )
      perror( "can't close input pipe!" );

    return ( status );
  }

  /* fork failure */
  perror( "fork()" );
  return ( -1 );
}

/* Sat Nov  8 14:24:53 PST 2003
 * Foneyop
 * read the command in from stdin */
char * get_command( char* command ) {
  char prompt[ 256 ];
#ifndef HAVE_LIBREADLINE
  int pos = 0;
#endif

  if ( command != NULL ) {
    free( command );
    command = NULL;
  }

  get_prompt( prompt, 255 );

#ifndef HAVE_LIBREADLINE

  if ( ( command = malloc( MAX_CMD ) ) == NULL )
    die( "no momory" );

  bzero( command, MAX_CMD );

  write( 1, prompt, strlen( prompt ) );

  while ( pos < MAX_CMD ) {
    read( 0, &command[ pos ], 1 );

    if ( command[ pos ] == 10 || command[ pos ] == '\0' )
      break;

    pos++;
  }

  command[ pos-- ] = '\0';
#endif

#ifdef HAVE_LIBREADLINE

  command = readline( prompt );

  if ( command && *command )
    add_history( command );

#endif

  return ( command );
}

/* Sat Nov  8 23:09:18 PST 2003
 * loop for ever and process commands */
void process_cmds( void ) {
  char *tokens  = NULL;
  char *command = NULL;
  char *value   = NULL;
  int error = 0;
  int num_tokens = 0;

  if ( ( tokens = malloc( MAX_TOKENS ) ) == NULL )
    perror( "malloc()" );

  /* read commands forever */
  while ( 1 ) {

next_command:
    command = get_command( command );

    /* display shell messages */
    if ( ( value = fu_get_conf( &g_conf, "shell_notify", 0 ) ) != NULL ) {
      fprintf( stderr, "%s", value );
      fu_get_conf( &g_conf, "shell_notify", 1 );
    }

    /* skip null commands */
    if ( command == NULL || command == ( char * ) - 1 ) {
      fu_exit( 0 );
    }

    /* tokenize the command */
    num_tokens = fu_tokenize( command, tokens, 0 );
    num_tokens = fu_expand( tokens, num_tokens );

    free( command );
    command = NULL;

    if ( num_tokens < 0 )
      goto next_command;

    /* received a quit or kill command */
    else if ( ( error = process_tokens( tokens, num_tokens ) ) == S_QUIT ) {
      free( tokens );
      exit( 0 );
    }

    /* try to exit gracefully from possibly boundry condition errors */
    else if ( error == S_INTERR ) {
      fprintf( stderr, "The command you entered caused an internal processing error\n" );
      fprintf( stderr, "Please report bugs to cmarsh@idahopower.com\n" );
      free( tokens );
      exit( -1 );
    }
  }
}


/* Tue Nov 25 11:55:40 MST 2003
 * take a parsed command and process it */
int process_tokens( char *tokens, int num_tokens ) {

  int tpos = 0;
  int optn = 0;
  int state = 0;
  int outpipe[ 2 ] = { 0, 1 };
  int inpipe = 0;
  int fu_flag = 0;
  int rstat = 99;
  char *argument = NULL;
  char *execmd = NULL;
  char *ptr = NULL;
  char buffer1[ MAXPATHLEN ];
  char buffer2[ MAXPATHLEN ];
  mode_t mode;
  glob_t globby;
  alias_cmd *alias = NULL;

  /* sometimes the glob structure has crap in it that can effect operation */
  bzero( &globby, sizeof( glob_t ) );

  /* iterate over each token, and go one round further */
  for ( tpos = 0; num_tokens >= 0; num_tokens-- ) {

    /* get the next argument */
    argument = &tokens[ tpos ];
    tpos += strlen( &tokens[ tpos ] ) + 1;

    /* don't run over the end of our buffer! */
    if ( tpos > MAX_TOKENS )
      return ( S_INTERR );

    /* don't do anything with null tokens */
    if ( argument != NULL && &argument[ 0 ] != '\0' )
      optn = get_state( argument );
    else
      continue;

#ifdef DEBUG
    printf( "%d - pos: [%d] got token: [%s] type: [%d]\n", num_tokens, tpos, argument, optn );
#endif

    /* user has quit */
    if ( optn == S_QUIT )
      return ( S_QUIT );

    /* no operation, ignore this keyword */
    if ( optn == S_NOP )
      continue;
    
    /* display shell options */
    if ( state == S_SET && optn != S_SETVALUE )
      fu_get_conf( &g_conf, NULL, 0 );

    /* change directory */
    if ( state == S_CD ) {
      getcwd( buffer1, MAXPATHLEN );
      fu_add_value( &g_conf, "OWD", buffer1 );

      if( strlen( argument ) < 1 || strcmp( argument, "~" ) == 0 )
        argument = fu_get_conf( &g_conf, "home", 0 );
      if ( chdir( ( char* ) argument ) != 0 )
        perror( "chdir()" );

      getcwd( buffer1, MAXPATHLEN );
      fu_add_value( &g_conf, "PWD", buffer1 );
    }

    /* don't wait on this process */
    else if ( optn == S_NOWAIT ) {
      fu_flag |= F_BACK;
      optn = S_LASTOPT;
    }

    /* print the current working directory */
    else if ( state == S_PWD ) {
      bzero( buffer1, MAXPATHLEN );

      if ( getcwd( buffer1, MAXPATHLEN ) == NULL )
        perror( "getwd()" );
      else
        printf( "%s\n", buffer1 );
    }

    /* set the umask */
    else if ( state == S_UMASK && optn != S_LASTOPT && optn != S_LASTCMD ) {
      sscanf( argument, "%04o", (mode_t *)&mode );
      umask( mode );
    }

    /* get the umask */
    else if ( state == S_UMASK && ( optn == S_LASTOPT || optn == S_LASTCMD ) &&
    execmd == NULL ) {
      mode = umask( 0 );
      umask( mode );
      printf( "%04o\n", mode );
    }

    /* set an environment variable, don't change the shell! */
    else if ( state == S_EXPORT ) {
      if ( ( strncasecmp( argument, "shell", 5 ) != 0 ) &&
      ( strncasecmp( argument, "LD", 2 ) != 0 ) && 
      ( strncasecmp( argument, "DT", 2 ) != 0 ) ) {
        /* export a shell variable */
        if( ( ptr = fu_get_conf( &g_conf, argument, 0 ) ) != NULL ) {
          argument = strsep( &argument, "=" );
          snprintf( buffer1, MAXPATHLEN, "%s=%s", argument, ptr );
          unsetenv( argument );
          putenv( buffer1 );
        }
        /* export a new variable */
        else {
          ptr = strsep( &argument, "=" );
          snprintf( buffer1, MAXPATHLEN, "%s=%s", ptr, argument );
          unsetenv( ptr );
          putenv( buffer1 );
        }
      }
      else {
        syndenied( "secure environment variable" );
      }
    }

    /* UNset a shell variable */
    else if ( state == S_UNSET )
      fu_get_conf( &g_conf, argument, 1 );

    /* UNsetenv an environment variable */
    else if ( state == S_UNENV ) {
      if ( ( strncasecmp( argument, "shell", 5 ) != 0 ) &&
      ( strncasecmp( argument, "LD", 2 ) != 0 ) && 
      ( strncasecmp( argument, "DT", 2 ) != 0 ) ) {
        unsetenv( argument );
      }
      else {
        syndenied( "secure environment variable" );
      }
    }

    /* set a shell option */
    else if ( optn == S_SETVALUE ) {
      if( ( ptr = strsep( &argument, "=" ) ) == NULL )
        syndenied( argument );
      snprintf( buffer1, MAXPATHLEN, "%s", ptr );

      if( ( ptr = strsep( &argument, "=" ) ) == NULL )
        syndenied( argument );
      snprintf( buffer2, MAXPATHLEN, "%s", ptr );

      if( strncmp( buffer1, "fu", 2 ) != 0 )
        fu_add_value( &g_conf, buffer1, buffer2 );
    }

    /* redirect output to a pipe */
    else if ( optn == S_PIPE ) {
      if ( pipe( outpipe ) != 0 )
        die( "can't open pipe" );

      optn = S_LASTOPT;
      fu_flag |= F_PIPE;
    }

    /* open a file, and redirect output to the file */
    else if ( state == S_RDROVR ) {
      if ( ( outpipe[ 1 ] = open( argument,
                                  O_WRONLY | O_CREAT, 0666 ) ) == -1 ) {
        perror( "open()" );
        optn = S_LASTCMD;
        execmd = NULL;
      }
    }

    /* open a file, and redirect output to the file */
    else if ( state == S_RDRAPD ) {
      if ( ( outpipe[ 1 ] = open( argument,
                                  O_APPEND | O_WRONLY | O_CREAT, 0666 ) ) == -1 ) {
        perror( "open()" );
        optn = S_LASTCMD;
        execmd = NULL;
      }
    }

    /* open a file, and redirect input from the file */
    else if ( state == S_RDRIN ) {
      if ( ( inpipe = open( argument, O_RDONLY ) ) == -1 ) {
        perror( argument );
        optn = S_LASTCMD;
        execmd = NULL;
      }
    }

    else if ( state == S_ALIAS ) {
      if ( ( alias = ( alias_cmd * ) malloc( sizeof( alias_cmd ) ) ) == NULL )
        die( "malloc()" );
      bzero( alias, sizeof( alias_cmd ) );
      strncpy( alias->name, argument, MAXPATHLEN );
      fu_flag |= F_ALIAS;
    }

    else if ( state == S_UNALIAS ) {
      if( fu_unalias( argument ) != 0 )
        fprintf( stderr, "Alias [%s] does not exist\n", argument );
    }

    /* first unknown argument in this command, set it to command name */
    else if ( optn == S_UNKNOWN && execmd == NULL ) {

      if ( fu_flag & F_ALIAS ) {
        if ( ( ptr = find_alias( argument, &g_alias ) ) != NULL )
          argument = ptr;

        strncpy( &alias->command[ alias->len ], argument, MAX_CMD - alias->len );
        alias->len += strlen( argument );
        alias->command[ alias->len ] = ' ';
        alias->len++;
      } else if ( fu_glob( &globby, argument ) == -1 ) {
        optn = S_LASTCMD;
        execmd = NULL;
      } else {
        execmd = argument;
      }
    }

    /* we already have a command, this is an argument */
    else if ( optn == S_UNKNOWN && execmd != NULL ) {
      if ( fu_glob( &globby, argument ) == -1 ) {
        optn = S_LASTCMD;
        execmd = NULL;
      }
    }

    /* do a logical AND with the return value */
    else if ( optn == S_AND ) {
      if ( ( fu_flag & F_HAVEBAND ) || ( fu_flag & F_HAVEBOR ) ||
           num_tokens < 1 ) {
        printf( "Invalid null command\n" );
        optn = S_LASTCMD;
        execmd = NULL;
      }
      fu_flag |= F_HAVEBAND;
      optn = S_LASTOPT;
    }

    /* do a logical OR with the return value */
    else if ( optn == S_OR ) {
      if ( ( fu_flag & F_HAVEBAND ) || ( fu_flag & F_HAVEBOR ) ||
           num_tokens < 1 ) {
        printf( "Invalid null command\n" );
        optn = S_LASTCMD;
        execmd = NULL;
      }
      fu_flag |= F_HAVEBOR;
      optn = S_LASTOPT;
    }

    /* this is the last option or command, execute our command */
    if ( ( optn == S_LASTOPT || optn == S_LASTCMD ) && execmd != NULL ) {

      rstat = fu_exec( execmd, globby.gl_pathv, outpipe[ 1 ], inpipe, fu_flag );
      /* pipe, set the input of the next command to the output of this */
      if ( fu_flag & F_PIPE )
        inpipe = outpipe[ 0 ];
      /* reset output and input to defaults */
      else
        outpipe[ 0 ] = inpipe = outpipe[ 1 ] = 0;

      /* command returned, reset all options */
      globfree( &globby );
      bzero( &globby, sizeof( globby ) );
      state = 0;
      outpipe[ 1 ] = 1;
      execmd = NULL;

      /* check for logical OR and logical AND */
      if ( ( fu_flag & F_HAVEBOR ) && ! fu_get_conf( &g_conf, "status", 0 ) ) {
        fu_flag = 0;
        fu_flag |= F_NOEXEC;
      } 
      else if ( ( fu_flag & F_HAVEBAND ) && 
                  fu_get_conf( &g_conf, "status", 0 ) ) {
        fu_flag = 0;
        fu_flag |= F_NOEXEC;
      } else
        fu_flag = 0;

      /* if canmulti is not set, then this is the last command */
      if ( fu_get_conf( &g_seconf, "fu_multi_command", 0 ) )
        optn = S_LASTOPT;
      else
        optn = S_LASTCMD;
    }

    /* this is the end of an alias command */
    else if ( ( optn == S_LASTOPT || optn == S_LASTCMD ) && ( fu_flag & F_ALIAS ) ) {
      if ( strlen( alias->command ) < 1 )
        fu_show_alias( &g_alias );
      else {
        alias->command[ alias->len - 1 ] = '\0';
        list_ins_next( &g_alias, list_tail( &g_alias ), alias );
      }
      globfree( &globby );
      bzero( &globby, sizeof( globby ) );
      fu_flag = state = 0;
      outpipe[ 1 ] = 1;
      execmd = NULL;
    }

    /* this is the last command, break out */
    if ( optn == S_LASTCMD ) {
      break;
    }

    /* set the current state to this command type */
    state = optn;
  }
  return( 0 );
}


/* Sat Nov  8 22:43:53 PST 2003
* set stuff up and call main loop */
int main( int argc, char **argv ) {
  char buffer[ BUFFMEDM ];
  char *conf = NULL;
  char *home = NULL;
  int  error = 0;
  uid_t uid  = 0;

   /* Initialize our linked lists */
  bzero( buffer, BUFFMEDM );
  list_init( ( List * ) &g_procs,  ( void * ) list_destroy );
  list_init( ( List * ) &g_alias,  ( void * ) list_destroy );
  list_init( ( List * ) &g_db,     ( void * ) list_destroy );
  list_init( ( List * ) &g_exec,   ( void * ) list_destroy );
  list_init( ( List * ) &g_conf,   ( void * ) list_destroy );
  list_init( ( List * ) &g_seconf, ( void * ) list_destroy );
  list_init( ( List * ) &g_owner,  ( void * ) list_destroy );
  list_init( ( List * ) &g_group,  ( void * ) list_destroy );

  uid = geteuid();
  sprintf(buffer, "%d", uid); 
  fu_add_value( &g_seconf, "euid", buffer );
  uid = getegid();
  sprintf(buffer, "%d", uid); 
  fu_add_value( &g_seconf, "egid", buffer );

  /* drop setuid */
  uid = getuid();
  error = seteuid( uid );
  //printf("euid: [%d] uid: [%d] seteuid err [%d]\n", euid, uid, error);
  //error = seteuid( euid );
  //printf("euid: [%d] uid: [%d] seteuid2 err [%d]\n", euid, uid, error);
  //die("crap!");
  uid = getgid();
  setegid( uid );

  /* set hard coded default options */
  fu_conf_default( &g_conf );
  home = fu_get_conf( &g_conf, "user", 0 );

   /* always catch terminated child processes */
  signal( SIGCHLD, &catch_chld );


  /* always read the security configuration files */
  snprintf( buffer, BUFFMEDM, "%s/fush.db", _CONFIGDIR_ );
  read_db( buffer );
  snprintf( buffer, BUFFMEDM, "%s/%s.security", _CONFIGDIR_, home );
  error = read_security( buffer );
  snprintf( buffer, BUFFMEDM, "%s/%s.security", _CONFIGDIR_, home );
  if ( ( error == -1 ) && ( read_security( buffer ) == -1 ) )
    die( "user security not set" );

  /* open a connection to syslog */
  if( !(conf = fu_get_conf( &g_seconf, "fu_log_facility", 0 ) ) )
		error = 128;
	else
		error = atoi( conf );

  if( fu_get_conf( &g_seconf, "fu_syslog", 0 ) )
		openlog( home, LOG_NDELAY | LOG_PID, error );

  /* check if this is an interactive shell, setup some stuff */
  if ( isatty( 0 ) && isatty( 1 ) ) {
    fu_add_value( &g_conf, "interactive", "yes" );
		/* dup the terminal for interactive shells */
		fu_dup_tty();

    signal( SIGINT,  &catch_int );
    signal( SIGHUP,  SIG_IGN );
    /* interactive access not allowed, process command line arguments, exit */
    if( !fu_get_conf( &g_seconf, "fu_allow_interactive", 0 ) ) {
      fu_do_opts( argc, argv );
      syndenied( "interactive access" );
      exit( -1 );
    }
  }

  /* process command line arguments */
  fu_do_opts( argc, argv );
  /* process the resource scripts */
  home = fu_get_conf( &g_conf, "home", 0 );
  if ( ( conf = fu_get_conf( &g_seconf, "fu_system_rc", 0 ) ) ) {
    proc_resource( conf );
  }
  if ( ( conf = fu_get_conf( &g_seconf, "fu_home_rc", 0 ) ) ) {
    snprintf( buffer, BUFFMEDM, "%s/%s", home, conf );
    proc_resource( buffer );
  }
  if ( fu_get_conf( &g_seconf, "fu_use_systemrc", 0 ) ) {
    snprintf( buffer, BUFFMEDM, "%s/fushrc", _CONFIGDIR_ );
    proc_resource( buffer );
  }
  if ( fu_get_conf( &g_seconf, "fu_use_homerc", 0 ) ) {
    snprintf( buffer, BUFFMEDM, "%s/.fushrc", home );
    proc_resource( buffer );
  }

  /* get commands from stdin and exec them */
  process_cmds( );

  /* control not reached */
  fu_exit( -1 );
  return ( -1 );
}
